﻿using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Turing
{
    public class Czwórka
    {
        public char StanGłowicy { get; private set; }
        public char StanKomórki { get; private set; }
        public char NowyStanKomórki { get; private set; }
        public char NowyStanGłowicy { get; private set; }

        public override string ToString()
        {
            return "" + StanGłowicy + StanKomórki + NowyStanKomórki + NowyStanGłowicy;
        }

        public Czwórka(char stanGłowicy, char stanKomórki, char nowyStanKomórki, char nowyStanGłowicy)
        {
            StanGłowicy = stanGłowicy;
            StanKomórki = stanKomórki;
            NowyStanKomórki = nowyStanKomórki;
            NowyStanGłowicy = nowyStanGłowicy;

            sprawdźStrukturę(this);
        }

        //pseudokonstruktor
        public static Czwórka Parsuj(string linia)
        {
            if (linia.Length != 4) throw new TuringException("Nieprawidłowa długość linii");
            Czwórka czwórka = new Czwórka(linia[0], linia[1], linia[2], linia[3]);
            //TODO: czy sprawdzać poprawność
            return czwórka;
        }

        private static bool sprawdźStrukturę(Czwórka czwórka)
        {
            bool czyLitery =
                char.IsLetter(czwórka.StanGłowicy) &&
                char.IsLetter(czwórka.StanKomórki) &&
                char.IsLetter(czwórka.NowyStanKomórki) &&
                char.IsLetter(czwórka.NowyStanGłowicy);
            bool czyPrawidłoweWielkości =
                char.IsLower(czwórka.StanGłowicy) &&
                char.IsUpper(czwórka.StanKomórki) &&
                char.IsUpper(czwórka.NowyStanKomórki) &&
                char.IsLower(czwórka.NowyStanGłowicy);
            return czyLitery && czyPrawidłoweWielkości;
        }

        public bool CzyPasuje(char stanGłowicy, char stanKomórki) //TODO: lepsza nazwa
        {
            return (StanGłowicy == stanGłowicy) && (StanKomórki == stanKomórki);
        }

        public class PorównywaczCzwórek : IEqualityComparer<Czwórka>
        {
            public bool Equals(Czwórka czwórka, Czwórka linia)
            {
                //return (czwórka.StanGłowicy == linia.StanGłowicy) && (czwórka.StanKomórki == linia.StanKomórki);
                return czwórka.CzyPasuje(linia.StanGłowicy, linia.StanKomórki);
            }

            public int GetHashCode([DisallowNull] Czwórka czwórka)
            {
                return czwórka.StanGłowicy ^ czwórka.StanKomórki ^ czwórka.NowyStanKomórki ^ czwórka.NowyStanGłowicy;
            }
        }
    }

    public class TuringException : Exception
    {
        public TuringException(string message = null, Exception innerException = null) 
            :base(message,innerException)
        {
        }
    }

    public class Kod
    {
        private List<Czwórka> program = new List<Czwórka>();

        private bool sprawdźCzyBezPowtórzeń(Czwórka czwórka)
        {
            foreach (Czwórka linia in program)
            {
                if ((czwórka.StanGłowicy == linia.StanGłowicy) && (czwórka.StanKomórki == linia.StanKomórki)) return false;
            }
            return true;
        }

        private static Czwórka.PorównywaczCzwórek porównywacz = new Czwórka.PorównywaczCzwórek();

        private bool sprawdźCzyBezPowtórzeńWWersjiPDawida(Czwórka czwórka)
        {
            return !program.Contains(czwórka, porównywacz);
        }

        //public bool Dodaj((char StanGłowicy, char StanKomórki, char NowyStanGłowicy, char NowyStanKomórki) czwórka)
        public void Dodaj(Czwórka czwórka)
        {
            //if (!sprawdźStrukturę(czwórka)) return false;
            if (!sprawdźCzyBezPowtórzeń(czwórka)) throw new TuringException("Powtórzona linia kodu: '" + czwórka.ToString() + "'");
            //if (!sprawdźCzyBezPowtórzeńWWersjiPDawida(czwórka)) return false;
            program.Add(czwórka);
        }

        public static Kod Wczytaj(string ścieżkaPliku)
        {
            try
            {
                string[] linie = File.ReadAllLines(ścieżkaPliku);
                Kod kod = new Kod();
                foreach (string linia in linie)
                {
                    if (linia.StartsWith("#")) continue;
                    Czwórka czwórka = Czwórka.Parsuj(linia);
                    kod.Dodaj(czwórka);
                }
                return kod;
            }
            catch(TuringException exc)
            {
                throw exc;
            }
            catch(Exception exc)
            {
                throw new TuringException("Inny wyjątek", exc);
            }            
        }

        public int LiczbaLinii
        {
            get
            {
                return program.Count;
            }
        }

        public Czwórka? ZnajdźPasującąCzwórkę(char stanGłowicy, char stanKomórki)
        {
            return program.FirstOrDefault(cz => cz.CzyPasuje(stanGłowicy, stanKomórki), null);
        }
    }
}
